# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import abc
import functools
import inspect
import logging
import pickle
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from flask import Flask, Request, Response, abort, request
import requests
from werkzeug.exceptions import HTTPException

from .negotiation import Formatter as FormatterBase
from .negotiation import Negotiator as NegotiatorBase
from .negotiation import negotiate as _negotiate
from .serializers import (
    exception_to_dict,
    json_dumps,
    json_loads,
    msgpack_dumps,
    msgpack_loads,
)
from .serializers import decode_response
from .serializers import encode_data_client as encode_data

logger = logging.getLogger(__name__)


# support for content negotiation


class Negotiator(NegotiatorBase):
    def best_mimetype(self):
        return request.accept_mimetypes.best_match(
            self.accept_mimetypes, "application/json"
        )

    def _abort(self, status_code, err=None):
        return abort(status_code, err)


def negotiate(formatter_cls, *args, **kwargs):
    return _negotiate(Negotiator, formatter_cls, *args, **kwargs)


class Formatter(FormatterBase):
    def _make_response(self, body, content_type):
        return Response(body, content_type=content_type)

    def configure(self, extra_encoders=None):
        self.extra_encoders = extra_encoders


class JSONFormatter(Formatter):
    format = "json"
    mimetypes = ["application/json"]

    def render(self, obj):
        return json_dumps(obj, extra_encoders=self.extra_encoders)


class MsgpackFormatter(Formatter):
    format = "msgpack"
    mimetypes = ["application/x-msgpack"]

    def render(self, obj):
        return msgpack_dumps(obj, extra_encoders=self.extra_encoders)


# base API classes


class RemoteException(Exception):
    """raised when remote returned an out-of-band failure notification, e.g., as a
    HTTP status code or serialized exception

    Attributes:
        response: HTTP response corresponding to the failure

    """

    def __init__(
        self,
        payload: Optional[Any] = None,
        response: Optional[requests.Response] = None,
    ):
        if payload is not None:
            super().__init__(payload)
        else:
            super().__init__()
        self.response = response

    def __str__(self):
        if (
            self.args
            and isinstance(self.args[0], dict)
            and "type" in self.args[0]
            and "args" in self.args[0]
        ):
            return (
                f"<RemoteException {self.response.status_code} "
                f'{self.args[0]["type"]}: {self.args[0]["args"]}>'
            )
        else:
            return super().__str__()


F = TypeVar("F", bound=Callable)


def remote_api_endpoint(path: str, method: str = "POST") -> Callable[[F], F]:
    def dec(f: F) -> F:
        f._endpoint_path = path  # type: ignore
        f._method = method  # type: ignore
        return f

    return dec


class APIError(Exception):
    """API Error"""

    def __str__(self):
        return "An unexpected error occurred in the backend: {}".format(self.args)


class MetaRPCClient(type):
    """Metaclass for RPCClient, which adds a method for each endpoint
    of the database it is designed to access.

    See for example :class:`swh.indexer.storage.api.client.RemoteStorage`"""

    def __new__(cls, name, bases, attributes):
        # For each method wrapped with @remote_api_endpoint in an API backend
        # (eg. :class:`swh.indexer.storage.IndexerStorage`), add a new
        # method in RemoteStorage, with the same documentation.
        #
        # Note that, despite the usage of decorator magic (eg. functools.wrap),
        # this never actually calls an IndexerStorage method.
        backend_class = attributes.get("backend_class", None)
        for base in bases:
            if backend_class is not None:
                break
            backend_class = getattr(base, "backend_class", None)
        if backend_class:
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    cls.__add_endpoint(meth_name, meth, attributes)
        return super().__new__(cls, name, bases, attributes)

    @staticmethod
    def __add_endpoint(meth_name, meth, attributes):
        wrapped_meth = inspect.unwrap(meth)

        @functools.wraps(meth)  # Copy signature and doc
        def meth_(*args, **kwargs):
            # Match arguments and parameters
            post_data = inspect.getcallargs(wrapped_meth, *args, **kwargs)

            # Remove arguments that should not be passed
            self = post_data.pop("self")
            post_data.pop("cur", None)
            post_data.pop("db", None)

            # Send the request.
            return self.post(meth._endpoint_path, post_data)

        if meth_name not in attributes:
            attributes[meth_name] = meth_


class RPCClient(metaclass=MetaRPCClient):
    """Proxy to an internal SWH RPC

    """

    backend_class = None  # type: ClassVar[Optional[type]]
    """For each method of `backend_class` decorated with
    :func:`remote_api_endpoint`, a method with the same prototype and
    docstring will be added to this class. Calls to this new method will
    be translated into HTTP requests to a remote server.

    This backend class will never be instantiated, it only serves as
    a template."""

    api_exception = APIError  # type: ClassVar[Type[Exception]]
    """The exception class to raise in case of communication error with
    the server."""

    reraise_exceptions: ClassVar[List[Type[Exception]]] = []
    """On server errors, if any of the exception classes in this list
    has the same name as the error name, then the exception will
    be instantiated and raised instead of a generic RemoteException."""

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(
        self,
        url,
        api_exception=None,
        timeout=None,
        chunk_size=4096,
        reraise_exceptions=None,
        **kwargs,
    ):
        if api_exception:
            self.api_exception = api_exception
        if reraise_exceptions:
            self.reraise_exceptions = reraise_exceptions
        base_url = url if url.endswith("/") else url + "/"
        self.url = base_url
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(
            max_retries=kwargs.get("max_retries", 3),
            pool_connections=kwargs.get("pool_connections", 20),
            pool_maxsize=kwargs.get("pool_maxsize", 100),
        )
        self.session.mount(self.url, adapter)

        self.timeout = timeout
        self.chunk_size = chunk_size

    def _url(self, endpoint):
        return "%s%s" % (self.url, endpoint)

    def raw_verb(self, verb, endpoint, **opts):
        if "chunk_size" in opts:
            # if the chunk_size argument has been passed, consider the user
            # also wants stream=True, otherwise, what's the point.
            opts["stream"] = True
        if self.timeout and "timeout" not in opts:
            opts["timeout"] = self.timeout
        try:
            return getattr(self.session, verb)(self._url(endpoint), **opts)
        except requests.exceptions.ConnectionError as e:
            raise self.api_exception(e)

    def post(self, endpoint, data, **opts):
        if isinstance(data, (abc.Iterator, abc.Generator)):
            data = (self._encode_data(x) for x in data)
        else:
            data = self._encode_data(data)
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "post",
            endpoint,
            data=data,
            headers={
                "content-type": "application/x-msgpack",
                "accept": "application/x-msgpack",
            },
            **opts,
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def _encode_data(self, data):
        return encode_data(data, extra_encoders=self.extra_type_encoders)

    post_stream = post

    def get(self, endpoint, **opts):
        chunk_size = opts.pop("chunk_size", self.chunk_size)
        response = self.raw_verb(
            "get", endpoint, headers={"accept": "application/x-msgpack"}, **opts
        )
        if opts.get("stream") or response.headers.get("transfer-encoding") == "chunked":
            self.raise_for_status(response)
            return response.iter_content(chunk_size)
        else:
            return self._decode_response(response)

    def get_stream(self, endpoint, **opts):
        return self.get(endpoint, stream=True, **opts)

    def raise_for_status(self, response) -> None:
        """check response HTTP status code and raise an exception if it denotes an
        error; do nothing otherwise

        """
        status_code = response.status_code
        status_class = response.status_code // 100

        if status_code == 404:
            raise RemoteException(payload="404 not found", response=response)

        exception = None

        # TODO: only old servers send pickled error; stop trying to unpickle
        # after they are all upgraded
        try:
            if status_class == 4:
                data = self._decode_response(response, check_status=False)
                if isinstance(data, dict):
                    # TODO: remove "exception" key check once all servers
                    # are using new schema
                    exc_data = data["exception"] if "exception" in data else data
                    for exc_type in self.reraise_exceptions:
                        if exc_type.__name__ == exc_data["type"]:
                            exception = exc_type(*exc_data["args"])
                            break
                    else:
                        exception = RemoteException(payload=exc_data, response=response)
                else:
                    exception = pickle.loads(data)

            elif status_class == 5:
                data = self._decode_response(response, check_status=False)
                if "exception_pickled" in data:
                    exception = pickle.loads(data["exception_pickled"])
                else:
                    # TODO: remove "exception" key check once all servers
                    # are using new schema
                    exc_data = data["exception"] if "exception" in data else data
                    exception = RemoteException(payload=exc_data, response=response)

        except (TypeError, pickle.UnpicklingError):
            raise RemoteException(payload=data, response=response)

        if exception:
            raise exception from None

        if status_class != 2:
            raise RemoteException(
                payload=f"API HTTP error: {status_code} {response.content}",
                response=response,
            )

    def _decode_response(self, response, check_status=True):
        if check_status:
            self.raise_for_status(response)
        return decode_response(response, extra_decoders=self.extra_type_decoders)

    def __repr__(self):
        return "<{} url={}>".format(self.__class__.__name__, self.url)


class BytesRequest(Request):
    """Request with proper escaping of arbitrary byte sequences."""

    encoding = "utf-8"
    encoding_errors = "surrogateescape"


ENCODERS: Dict[str, Callable[[Any], Union[bytes, str]]] = {
    "application/x-msgpack": msgpack_dumps,
    "application/json": json_dumps,
}


def encode_data_server(
    data, content_type="application/x-msgpack", extra_type_encoders=None
):
    encoded_data = ENCODERS[content_type](data, extra_encoders=extra_type_encoders)
    return Response(encoded_data, mimetype=content_type,)


def decode_request(request, extra_decoders=None):
    content_type = request.mimetype
    data = request.get_data()
    if not data:
        return {}

    if content_type == "application/x-msgpack":
        r = msgpack_loads(data, extra_decoders=extra_decoders)
    elif content_type == "application/json":
        # XXX this .decode() is needed for py35.
        # Should not be needed any more with py37
        r = json_loads(data.decode("utf-8"), extra_decoders=extra_decoders)
    else:
        raise ValueError("Wrong content type `%s` for API request" % content_type)

    return r


def error_handler(exception, encoder, status_code=500):
    logging.exception(exception)
    response = encoder(exception_to_dict(exception))
    if isinstance(exception, HTTPException):
        response.status_code = exception.code
    else:
        # TODO: differentiate between server errors and client errors
        response.status_code = status_code
    return response


class RPCServerApp(Flask):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class:
        The class of the backend, which will be analyzed to look
        for API endpoints.
    :param Optional[Callable[[], backend_class]] backend_factory:
        A function with no argument that returns an instance of
        `backend_class`. If unset, defaults to calling `backend_class`
        constructor directly.
    """

    request_class = BytesRequest

    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(self, *args, backend_class=None, backend_factory=None, **kwargs):
        super().__init__(*args, **kwargs)
        if backend_class is None and backend_factory is not None:
            raise ValueError(
                "backend_factory should only be provided if backend_class is"
            )

        self.backend_class = backend_class
        if backend_class is not None:
            backend_factory = backend_factory or backend_class
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    self.__add_endpoint(meth_name, meth, backend_factory)

    def __add_endpoint(self, meth_name, meth, backend_factory):
        from flask import request

        @self.route("/" + meth._endpoint_path, methods=["POST"])
        @negotiate(MsgpackFormatter, extra_encoders=self.extra_type_encoders)
        @negotiate(JSONFormatter, extra_encoders=self.extra_type_encoders)
        @functools.wraps(meth)  # Copy signature and doc
        def _f():
            # Call the actual code
            obj_meth = getattr(backend_factory(), meth_name)
            kw = decode_request(request, extra_decoders=self.extra_type_decoders)
            return obj_meth(**kw)



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json

import msgpack
import pytest

from swh.core.api.asynchronous import (
    Response,
    RPCServerApp,
    decode_data,
    decode_request,
    encode_msgpack,
)
from swh.core.api.serializers import SWHJSONEncoder, json_dumps, msgpack_dumps

pytest_plugins = ["aiohttp.pytest_plugin", "pytester"]


class TestServerException(Exception):
    pass


class TestClientError(Exception):
    pass


async def root(request):
    return Response("toor")


STRUCT = {
    "txt": "something stupid",
    # 'date': datetime.date(2019, 6, 9),  # not supported
    "datetime": datetime.datetime(2019, 6, 9, 10, 12, tzinfo=datetime.timezone.utc),
    "timedelta": datetime.timedelta(days=-2, hours=3),
    "int": 42,
    "float": 3.14,
    "subdata": {
        "int": 42,
        "datetime": datetime.datetime(
            2019, 6, 10, 11, 12, tzinfo=datetime.timezone.utc
        ),
    },
    "list": [
        42,
        datetime.datetime(2019, 9, 10, 11, 12, tzinfo=datetime.timezone.utc),
        "ok",
    ],
}


async def struct(request):
    return Response(STRUCT)


async def echo(request):
    data = await decode_request(request)
    return Response(data)


async def server_exception(request):
    raise TestServerException()


async def client_error(request):
    raise TestClientError()


async def echo_no_nego(request):
    # let the content negotiation handle the serialization for us...
    data = await decode_request(request)
    ret = encode_msgpack(data)
    return ret


def check_mimetype(src, dst):
    src = src.split(";")[0].strip()
    dst = dst.split(";")[0].strip()
    assert src == dst


@pytest.fixture
def async_app():
    app = RPCServerApp()
    app.client_exception_classes = (TestClientError,)
    app.router.add_route("GET", "/", root)
    app.router.add_route("GET", "/struct", struct)
    app.router.add_route("POST", "/echo", echo)
    app.router.add_route("GET", "/server_exception", server_exception)
    app.router.add_route("GET", "/client_error", client_error)
    app.router.add_route("POST", "/echo-no-nego", echo_no_nego)
    return app


@pytest.fixture
def cli(async_app, aiohttp_client, loop):
    return loop.run_until_complete(aiohttp_client(async_app))


async def test_get_simple(cli) -> None:
    resp = await cli.get("/")
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    data = await resp.read()
    value = msgpack.unpackb(data, raw=False)
    assert value == "toor"


async def test_get_server_exception(cli) -> None:
    resp = await cli.get("/server_exception")
    assert resp.status == 500
    data = await resp.read()
    data = msgpack.unpackb(data, raw=False)
    assert data["type"] == "TestServerException"


async def test_get_client_error(cli) -> None:
    resp = await cli.get("/client_error")
    assert resp.status == 400
    data = await resp.read()
    data = msgpack.unpackb(data, raw=False)
    assert data["type"] == "TestClientError"


async def test_get_simple_nego(cli) -> None:
    for ctype in ("x-msgpack", "json"):
        resp = await cli.get("/", headers={"Accept": "application/%s" % ctype})
        assert resp.status == 200
        check_mimetype(resp.headers["Content-Type"], "application/%s" % ctype)
        assert (await decode_request(resp)) == "toor"


async def test_get_struct(cli) -> None:
    """Test returned structured from a simple GET data is OK"""
    resp = await cli.get("/struct")
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    assert (await decode_request(resp)) == STRUCT


async def test_get_struct_nego(cli) -> None:
    """Test returned structured from a simple GET data is OK"""
    for ctype in ("x-msgpack", "json"):
        resp = await cli.get("/struct", headers={"Accept": "application/%s" % ctype})
        assert resp.status == 200
        check_mimetype(resp.headers["Content-Type"], "application/%s" % ctype)
        assert (await decode_request(resp)) == STRUCT


async def test_post_struct_msgpack(cli) -> None:
    """Test that msgpack encoded posted struct data is returned as is"""
    # simple struct
    resp = await cli.post(
        "/echo",
        headers={"Content-Type": "application/x-msgpack"},
        data=msgpack_dumps({"toto": 42}),
    )
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    assert (await decode_request(resp)) == {"toto": 42}
    # complex struct
    resp = await cli.post(
        "/echo",
        headers={"Content-Type": "application/x-msgpack"},
        data=msgpack_dumps(STRUCT),
    )
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    assert (await decode_request(resp)) == STRUCT


async def test_post_struct_json(cli) -> None:
    """Test that json encoded posted struct data is returned as is"""
    resp = await cli.post(
        "/echo",
        headers={"Content-Type": "application/json"},
        data=json.dumps({"toto": 42}, cls=SWHJSONEncoder),
    )
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    assert (await decode_request(resp)) == {"toto": 42}

    resp = await cli.post(
        "/echo",
        headers={"Content-Type": "application/json"},
        data=json.dumps(STRUCT, cls=SWHJSONEncoder),
    )
    assert resp.status == 200
    check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
    # assert resp.headers['Content-Type'] == 'application/x-msgpack'
    assert (await decode_request(resp)) == STRUCT


async def test_post_struct_nego(cli) -> None:
    """Test that json encoded posted struct data is returned as is

    using content negotiation (accept json or msgpack).
    """
    for ctype in ("x-msgpack", "json"):
        resp = await cli.post(
            "/echo",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/%s" % ctype,
            },
            data=json.dumps(STRUCT, cls=SWHJSONEncoder),
        )
        assert resp.status == 200
        check_mimetype(resp.headers["Content-Type"], "application/%s" % ctype)
        assert (await decode_request(resp)) == STRUCT


async def test_post_struct_no_nego(cli) -> None:
    """Test that json encoded posted struct data is returned as msgpack

    when using non-negotiation-compatible handlers.
    """
    for ctype in ("x-msgpack", "json"):
        resp = await cli.post(
            "/echo-no-nego",
            headers={
                "Content-Type": "application/json",
                "Accept": "application/%s" % ctype,
            },
            data=json.dumps(STRUCT, cls=SWHJSONEncoder),
        )
        assert resp.status == 200
        check_mimetype(resp.headers["Content-Type"], "application/x-msgpack")
        assert (await decode_request(resp)) == STRUCT


def test_async_decode_data_failure():
    with pytest.raises(ValueError, match="Wrong content type"):
        decode_data("some-data", "unknown-content-type")


@pytest.mark.parametrize("data", [None, "", {}, []])
def test_async_decode_data_empty_cases(data):
    assert decode_data(data, "unknown-content-type") == {}


@pytest.mark.parametrize(
    "data,content_type,encode_data_fn",
    [
        ({"a": 1}, "application/json", json_dumps),
        ({"a": 1}, "application/x-msgpack", msgpack_dumps),
    ],
)
def test_async_decode_data_nominal(data, content_type, encode_data_fn):
    actual_data = decode_data(encode_data_fn(data), content_type)
    assert actual_data == data



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import OrderedDict
import functools
import logging
from typing import Callable, Dict, List, Optional, Tuple, Type, Union

import aiohttp.web
from aiohttp_utils import Response, negotiation
from deprecated import deprecated
import multidict

from .serializers import (
    exception_to_dict,
    json_dumps,
    json_loads,
    msgpack_dumps,
    msgpack_loads,
)


def encode_msgpack(data, **kwargs):
    return aiohttp.web.Response(
        body=msgpack_dumps(data),
        headers=multidict.MultiDict({"Content-Type": "application/x-msgpack"}),
        **kwargs,
    )


encode_data_server = Response


def render_msgpack(request, data, extra_encoders=None):
    return msgpack_dumps(data, extra_encoders=extra_encoders)


def render_json(request, data, extra_encoders=None):
    return json_dumps(data, extra_encoders=extra_encoders)


def decode_data(data, content_type, extra_decoders=None):
    """Decode data according to content type, eventually using some extra decoders.

    """
    if not data:
        return {}
    if content_type == "application/x-msgpack":
        r = msgpack_loads(data, extra_decoders=extra_decoders)
    elif content_type == "application/json":
        r = json_loads(data, extra_decoders=extra_decoders)
    else:
        raise ValueError(f"Wrong content type `{content_type}` for API request")

    return r


async def decode_request(request, extra_decoders=None):
    """Decode asynchronously the request

    """
    data = await request.read()
    return decode_data(data, request.content_type, extra_decoders=extra_decoders)


async def error_middleware(app, handler):
    async def middleware_handler(request):
        try:
            return await handler(request)
        except Exception as e:
            if isinstance(e, aiohttp.web.HTTPException):
                raise
            logging.exception(e)
            res = exception_to_dict(e)
            if isinstance(e, app.client_exception_classes):
                status = 400
            else:
                status = 500
            return encode_data_server(res, status=status)

    return middleware_handler


class RPCServerApp(aiohttp.web.Application):
    """For each endpoint of the given `backend_class`, tells app.route to call
    a function that decodes the request and sends it to the backend object
    provided by the factory.

    :param Any backend_class:
        The class of the backend, which will be analyzed to look
        for API endpoints.
    :param Optional[Callable[[], backend_class]] backend_factory:
        A function with no argument that returns an instance of
        `backend_class`. If unset, defaults to calling `backend_class`
        constructor directly.
    """

    client_exception_classes: Tuple[Type[Exception], ...] = ()
    """Exceptions that should be handled as a client error (eg. object not
    found, invalid argument)"""
    extra_type_encoders: List[Tuple[type, str, Callable]] = []
    """Value of `extra_encoders` passed to `json_dumps` or `msgpack_dumps`
    to be able to serialize more object types."""
    extra_type_decoders: Dict[str, Callable] = {}
    """Value of `extra_decoders` passed to `json_loads` or `msgpack_loads`
    to be able to deserialize more object types."""

    def __init__(
        self,
        app_name: Optional[str] = None,
        backend_class: Optional[Callable] = None,
        backend_factory: Optional[Union[Callable, str]] = None,
        middlewares=(),
        **kwargs,
    ):
        nego_middleware = negotiation.negotiation_middleware(
            renderers=self._renderers(), force_rendering=True
        )
        middlewares = (nego_middleware, error_middleware,) + middlewares
        super().__init__(middlewares=middlewares, **kwargs)

        # swh decorations starts here
        self.app_name = app_name
        if backend_class is None and backend_factory is not None:
            raise ValueError(
                "backend_factory should only be provided if backend_class is"
            )
        self.backend_class = backend_class
        if backend_class is not None:
            backend_factory = backend_factory or backend_class
            for (meth_name, meth) in backend_class.__dict__.items():
                if hasattr(meth, "_endpoint_path"):
                    path = meth._endpoint_path
                    http_method = meth._method
                    path = path if path.startswith("/") else f"/{path}"
                    self.router.add_route(
                        http_method,
                        path,
                        self._endpoint(meth_name, meth, backend_factory),
                    )

    def _renderers(self):
        """Return an ordered list of renderers in order of increasing desirability (!)
        See mimetype.best_match() docstring

        """
        return OrderedDict(
            [
                (
                    "application/json",
                    lambda request, data: render_json(
                        request, data, extra_encoders=self.extra_type_encoders
                    ),
                ),
                (
                    "application/x-msgpack",
                    lambda request, data: render_msgpack(
                        request, data, extra_encoders=self.extra_type_encoders
                    ),
                ),
            ]
        )

    def _endpoint(self, meth_name, meth, backend_factory):
        """Create endpoint out of the method `meth`.

        """

        @functools.wraps(meth)  # Copy signature and doc
        async def decorated_meth(request, *args, **kwargs):
            obj_meth = getattr(backend_factory(), meth_name)
            data = await request.read()
            kw = decode_data(
                data, request.content_type, extra_decoders=self.extra_type_decoders
            )
            result = obj_meth(**kw)
            return encode_data_server(result)

        return decorated_meth


@deprecated(version="0.0.64", reason="Use the RPCServerApp instead")
class SWHRemoteAPI(RPCServerApp):
    pass



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import logging.config
from typing import Optional
import warnings

import click
import pkg_resources

LOG_LEVEL_NAMES = ["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

logger = logging.getLogger(__name__)


class AliasedGroup(click.Group):
    """A simple Group that supports command aliases, as well as notes related to
    options"""

    def __init__(self, name=None, commands=None, **attrs):
        self.option_notes = attrs.pop("option_notes", None)
        self.aliases = {}
        super().__init__(name, commands, **attrs)

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, self.aliases.get(cmd_name, cmd_name))

    def add_alias(self, name, alias):
        if not isinstance(name, str):
            name = name.name
        self.aliases[alias] = name

    def format_options(self, ctx, formatter):
        click.Command.format_options(self, ctx, formatter)
        if self.option_notes:
            with formatter.section("Notes"):
                formatter.write_text(self.option_notes)
        self.format_commands(ctx, formatter)


def clean_exit_on_signal(signal, frame):
    """Raise a SystemExit exception to let command-line clients wind themselves
    down on exit"""
    raise SystemExit(0)


def validate_loglevel_params(ctx, param, value):
    """Validate the --log-level parameters, with multiple values"""
    if value is None:
        return None
    return [validate_loglevel(ctx, param, v) for v in value]


def validate_loglevel(ctx, param, value):
    """Validate a single loglevel specification, of the form LOGLEVEL or
    module:LOGLEVEL."""
    if ":" in value:
        try:
            module, log_level = value.split(":")
        except ValueError:
            raise click.BadParameter(
                "Invalid log level specification `%s`, "
                "needs to be in format `module:LOGLEVEL`" % value
            )
    else:
        module = None
        log_level = value

    if log_level not in LOG_LEVEL_NAMES:
        raise click.BadParameter(
            "Log level %s unknown (in `%s`) needs to be one of %s"
            % (log_level, value, ", ".join(LOG_LEVEL_NAMES))
        )

    return (module, log_level)


@click.group(
    context_settings=CONTEXT_SETTINGS,
    cls=AliasedGroup,
    option_notes="""\
If both options are present, --log-level values will override the configuration
in --log-config.

The --log-config YAML must conform to the logging.config.dictConfig schema
documented at https://docs.python.org/3/library/logging.config.html.
""",
)
@click.option(
    "--log-level",
    "-l",
    "log_levels",
    default=None,
    callback=validate_loglevel_params,
    multiple=True,
    help=(
        "Log level (defaults to INFO). "
        "Can override the log level for a specific module, by using the "
        "`specific.module:LOGLEVEL` syntax (e.g. `--log-level swh.core:DEBUG` "
        "will enable DEBUG logging for swh.core)."
    ),
)
@click.option(
    "--log-config",
    default=None,
    type=click.File("r"),
    help="Python yaml logging configuration file.",
)
@click.option(
    "--sentry-dsn", default=None, help="DSN of the Sentry instance to report to"
)
@click.option(
    "--sentry-debug/--no-sentry-debug",
    default=False,
    hidden=True,
    help="Enable debugging of sentry",
)
@click.pass_context
def swh(ctx, log_levels, log_config, sentry_dsn, sentry_debug):
    """Command line interface for Software Heritage.
    """
    import signal

    import yaml

    from ..sentry import init_sentry

    signal.signal(signal.SIGTERM, clean_exit_on_signal)
    signal.signal(signal.SIGINT, clean_exit_on_signal)

    init_sentry(sentry_dsn, debug=sentry_debug)

    set_default_loglevel: Optional[str] = None

    if log_config:
        logging.config.dictConfig(yaml.safe_load(log_config.read()))
        set_default_loglevel = logging.root.getEffectiveLevel()

    if not log_levels:
        log_levels = []

    for module, log_level in log_levels:
        logger = logging.getLogger(module)
        log_level = logging.getLevelName(log_level)
        logger.setLevel(log_level)

        if module is None:
            set_default_loglevel = log_level

    if not set_default_loglevel:
        logging.root.setLevel("INFO")
        set_default_loglevel = "INFO"

    ctx.ensure_object(dict)
    ctx.obj["log_level"] = set_default_loglevel


def main():
    # Even though swh() sets up logging, we need an earlier basic logging setup
    # for the next few logging statements
    logging.basicConfig()
    # load plugins that define cli sub commands
    for entry_point in pkg_resources.iter_entry_points("swh.cli.subcommands"):
        try:
            cmd = entry_point.load()
            if isinstance(cmd, click.BaseCommand):
                # for BW compat, auto add click commands
                warnings.warn(
                    f"{entry_point.name}: automagic addition of click commands "
                    f"to the main swh group is deprecated",
                    DeprecationWarning,
                )
                swh.add_command(cmd, name=entry_point.name)
            # otherwise it's expected to be a module which has been loaded
            # it's the responsibility of the click commands/groups in this
            # module to transitively have the main swh group as parent.
        except Exception as e:
            logger.warning("Could not load subcommand %s: %s", entry_point.name, str(e))

    return swh(auto_envvar_prefix="SWH")


if __name__ == "__main__":
    main()
